# constants
TERM_CHAR = b'\n'


class CommandParser:
    def __init__(self, initial_data=b''):
        self.data        = initial_data.lstrip()

    def receive(self, data):
        self.data += data
        self.data  = self.data.lstrip()

    def next(self):
        # data?
        if not self.data:
            return None

        # term char?
        i = self.data.find(TERM_CHAR)
        if i == -1:
            return None

        # parse
        command   = self.data[:i].strip()
        self.data = self.data[i+1:].lstrip()
        return command or self.next_command()



from   .mixins        import ErrorQueueMixin
from   .basic_command import BasicCommand
import json


class ErrorsCommand(BasicCommand):
    # define
    command = "errors?"
    args    = {}

    def __init__(self, error_queue):
        super().__init__()
        self.error_queue = error_queue

    def code(self, devices, args):
        # process errors into strings
        prev_errors = [str(i) for i in self.error_queue]

        # clear queue
        self.error_queue.clear()

        # TODO: fix format
        return json.dumps(prev_errors)



from .device_factory_base     import DeviceFactoryBase
from instrument_server.errors import DeviceError
from pyvisa                   import ResourceManager
from pyvisa.resources         import Resource


# constants
DEFAULT_TIMEOUT_S = 5.0


class VisaFactory(DeviceFactoryBase):
    type = 'visa'

    @staticmethod
    def open(**settings):
        # resource_str?
        if 'resource_str' not in settings:
            message = "'resource_str' is missing in open()"
            self.raise_open_error(message)

        # settings
        resource_str = settings['resource_str']
        timeout_s    = settings.get('timeout_s', DEFAULT_TIMEOUT_S)

        manager = ResourceManager()
        device  = manager.open_resource(resource_str, open_timeout=timeout_s)
        device.time_ms = timeout_s
        return device

    @staticmethod
    def close(device):
        device.close()


# export
IS_DEVICE_PLUGIN = True
plugin           = VisaFactory



from   .mixins import RaiseErrorMixin
import re


class CommandParser(RaiseErrorMixin):
    def __init__(self, command, args={}):
        RaiseErrorMixin.__init__(self)
        self.command = command
        self.args    = args

    def is_match(self, command_bytes):
        regex = f'^{self.command}\\s*'.replace('?', r'\?').encode()
        matches = re.match(regex, command_bytes)
        return matches is not None

    def parse_args(self, command_bytes):
        values = command_bytes.strip().split()[1:]
        if len(values) < len(self.args):
            self.raise_error('too few arguments')
        if len(values) > len(self.args):
            self.raise_error('too many arguments')

        args = {}
        for name, type, value in zip(self.args.keys(), self.args.values(), values):
            if type:
                try:  # type(arg), catch Exception
                    typed_value = type(value)
                except Exception:
                    self.raise_error(f"'{value}' could not be converted to {type}")
                args[name] = typed_value
            else:
                # decode bytes to str
                args[name] = value.decode()
        return args



from .device_factory_base     import DeviceFactoryBase
from instrument_server.errors import DeviceError
from socket                   import socket


# constants
DEFAULT_PORT      = 5025
DEFAULT_TIMEOUT_S = 2.0


class SocketFactory(DeviceFactoryBase):
    type = 'socket'

    def open(self, **settings):
        # is address?
        if 'address' not in settings:
            message = "'address' missing in open()"
            self.raise_open_error(message)

        # get settings
        address   = settings['address']
        port      = settings.get('port', DEFAULT_PORT)
        timeout_s = settings.get('timeout_s', DEFAULT_TIMEOUT_S)

        # new socket
        device = socket()

        # open and return
        try:  # socket.connect, catch Exception
            device.connect((address, port))
        except Exception as error:
            device.close()
            raise error

        # connected
        device.settimeout(timeout_s)
        return device

    def close(self, device):
        device.close()


# export
IS_DEVICE_PLUGIN = True
plugin           = SocketFactory



def to_bytes(result):
    if isinstance(result, bytes):
        return result
    if isinstance(result, str):
        # encode str to bytes
        return result.encode()

    # to "bytes str"
    return str(result).encode()



from socket import AddressFamily


def is_ipv4(socket):
    if socket.family == AddressFamily.AF_UNIX:
        # unix socket
        return False

    if socket.family == AddressFamily.AF_INET6:
        # IP v6
        return False

    return socket.family == AddressFamily.AF_INET



from .command_base   import CommandBase
from .command_parser import CommandParser


class BasicCommand(CommandBase):
    """Command Plugin base class with command, typed arg parsing"""

    command = ''
    """define command base"""

    args    = {}
    """define args, types"""

    def __init__(self, *args, **kwargs):
        CommandBase.__init__(self, *args, **kwargs)
        self.parser = CommandParser(self.command, self.args)

    def is_match(self, command_bytes):
        """returns `True` if self can execute `command_bytes`; False otherwise"""
        return self.parser.is_match(command_bytes)

    def execute(self, devices, command_bytes):
        """parses command and calls self.code"""
        # parse args
        args = self.parser.parse_args(command_bytes)

        # execute code
        return self.code(devices, args)

    def code(self, devices, args):
        """code to execute parsed command"""



from .basic_command  import BasicCommand
from .command_base   import CommandBase
from .errors_command import ErrorsCommand
from .quit_command   import QuitCommand


# export
__all__ = ['BasicCommand', 'CommandBase', 'ErrorsCommand', 'QuitCommand']



class ErrorQueueMixin:
    # TODO: fix error queue
    def __init__(self):
        if 'errors' not in self.state:
            self.state['errors'] = list()

    @property
    def error_queue(self):
        return self.state['errors']



from instrument_server.errors import DeviceError, OpenDeviceError


class RaiseErrorMixin:
    def raise_open_error(self, message):
        raise OpenDeviceError(f'{self.type} open error: {message}')

    def raise_error(self, message):
        raise DeviceError(f'{self.type} error: {message}')



from pathlib     import Path
from ruamel.yaml import YAML


# share global parser
# rt = round trip dump
# (aka pretty-print)
yaml = YAML(typ='rt')


def load_yaml(file):
    """read yaml from file"""
    file = Path(file)
    return yaml.load(file)


def save_yaml(data, file):
    """write yaml to file"""
    file = Path(file)
    return yaml.dump(data, file)



from .device_factory_base import DeviceFactoryBase
from .socket_factory      import SocketFactory
from .visa_factory        import VisaFactory


# export
__all__ = ['DeviceFactoryBase', 'SocketFactory', 'VisaFactory']



from .mixins import RaiseErrorMixin
from abc     import ABCMeta, abstractmethod


class CommandBase(RaiseErrorMixin, metaclass=ABCMeta):
    """Command Plugin base class"""

    @abstractmethod
    def is_match(self, command_bytes):
        """returns `True` if self can execute command; `False` otherwise"""

    @abstractmethod
    def execute(self, devices, command_bytes):
        """executes command (bytes) with devices"""



from .basic_command import BasicCommand
import sys


class QuitCommand(BasicCommand):
    """`quit`

    Shuts down the server
    """

    # definition
    command = 'quit'
    args    = {}

    # implementation
    def code(self, devices, args):
        sys.exit(0)


# export
IS_COMMAND_PLUGIN = True
plugin            = QuitCommand



from   .application import Application
from   .client      import Client
from   .helpers     import is_ipv4
import asyncio
from   asyncio      import CancelledError, start_server


# constants
BUFFER_SIZE  = 1024


class InstrumentServer:
    def __init__(self, plugins, devices):
        # tcp server
        self.shutdown = False
        self.clients  = []
        self.server   = None

        # app
        self.plugins     = plugins
        self.devices     = devices
        self.application = None

    @property
    def is_running(self):
        return self.server is not None

    async def serve_forever(self, address, port):
        try:
            assert not self.is_running
            self.application = Application(self.plugins, self.devices)
            await self.start_server(address, port)
            self.print_server_status()
            await self.server.serve_forever()
        except CancelledError:
            await self.close_clients()
            self.application.close_devices()
            print('shutdown', flush=True)

    async def stop(self):
        if self.server:
            # remove reference
            server = self.server
            self.server = None

            # close
            server.close()
            await server.wait_closed()

    def run(self, address, port):
        """starts a new tcp server and serves forever (blocking)"""
        assert not self.is_running
        try:
            asyncio.run(self.serve_forever(address, port))
        except SystemExit:
            pass
        except KeyboardInterrupt:
            pass


    # helpers

    def print_server_status(self):
        socket = self.server.sockets[0]
        if is_ipv4(socket):
            # print IPv6
            address, port = socket.getsockname()
            print(f'Running on {address}:{port}...', flush=True)
        else:
            # print IPv6
            print(f'Running on {socket.getsockname()}...', flush=True)

    # client helpers

    def add_client(self, client):
        self.clients.append(client)

    def remove_client(self, client):
        if client not in self.clients:
            return

        # find index
        index = self.clients.index(client)

        # delete by index
        del(self.clients[index])

    async def handle_new_client(self, reader, writer):
        client = Client(self.application, reader, writer)
        self.add_client(client)
        print(f'  connected:    {client}', flush=True)

        # for exception handling
        close_client = False
        shutdown     = False
        error        = None

        try:
            await client.serve_forever()
        except CancelledError:
            # handle client task cancelled
            close_client = True
            shutdown     = False
            # ignore error
        except SystemExit as e:
            # received quit command from client
            close_client = False
            shutdown     = True
            # ignore error
        except BaseException as e:
            # bad stuff happened
            close_client = True
            shutdown     = True
            error        = e

        # close
        self.remove_client(client)
        if close_client:
            await client.close()

        print(f'  disconnected: {client}', flush=True)

        # shut down server?
        if shutdown:
            await self.stop()

        if error:
            raise error

    async def close_clients(self):
        for client in self.clients:
            await client.close()
        self.clients.clear()

    # tcp socket helpers

    async def start_server(self, address, port):
        """returns a new asyncio tcp server coroutine"""
        assert not self.is_running
        self.shutdown = False
        self.server   = await start_server(self.handle_new_client, address, port)
        return self.server



from .command_parser import CommandParser
from .helpers        import to_bytes


# constants
BUFFER_SIZE = 1024


class Client:
    def __init__(self, application, reader, writer):
        self.reader      = reader
        self.writer      = writer
        self.application = application
        self.command_parser = CommandParser()

    @property
    def address(self):
        address, _ = self.writer.get_extra_info('peername')
        return address

    @property
    def port(self):
        _, port = self.writer.get_extra_info('peername')
        return port

    def __repr__(self):
        return f'Client({self.address}, {self.port})'

    async def serve_forever(self):
        while data := await self.read():
            self.command_parser.receive(data)
            while command := self.command_parser.next():
                self.execute(command)

    # helpers

    def execute(self, command):
        result = self.application.execute(command)
        if result is not None:
            # query; send result
            self.writer.write(to_bytes(result))

    async def close(self):
        if not self.writer.is_closing:
            self.writer.close()
        await self.writer.wait_closed()

    async def read(self):
        return await self.reader.read(BUFFER_SIZE)



from .mixins import RaiseErrorMixin
from abc     import ABCMeta, abstractmethod


class DeviceFactoryBase(RaiseErrorMixin, metaclass=ABCMeta):
    @property
    @abstractmethod
    def type(self):
        """device type for reference in devices list"""

    @abstractmethod
    def open(self, **settings):
        """return opened device with `settings`"""

    def close(self, device):
        """close `device`"""
        # device.close()



from .helpers         import ellipsis_bytes
from ..commands       import ErrorsCommand
from ..plugin_manager import PluginManager
from ..errors         import CommandNotFoundError, OpenDeviceError


class Application:
    def __init__(self, plugins, devices):
        # init plugins
        self.plugins          = PluginManager()
        self.device_factories = dict()
        self.commands         = []

        # init app state
        self.devices      = {}
        self.device_types = {}
        self.errors       = []  # TODO: fix error queue

        # add error handling
        errors_command = ErrorsCommand(self.errors)
        self.commands.append(errors_command)

        # built-ins
        plugins.append('instrument_server.commands.quit_command')
        plugins.append('instrument_server.devices.socket_factory')
        plugins.append('instrument_server.devices.visa_factory')

        # start
        self.load_plugins(plugins)
        try:  # open devices, catch BaseException
            self.open_devices(devices)
        except BaseException as error:
            self.close_devices()
            raise error

    def __del__(self):
        self.close_devices()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_devices()

    def execute(self, command_bytes):
        for command in self.commands:
            if command.is_match(command_bytes):
                try:  # execute command, catch Exception
                    return command.execute(self.devices, command_bytes)
                except Exception as error:
                    # log non-exiting error
                    print(error, flush=True)
                    self.errors.append(error)
                    return

        # error: command not found
        message = f"command '{ellipsis_bytes(command_bytes)}' not found"
        error   = CommandNotFoundError(message)
        print(error, flush=True)
        self.errors.append(error)

    # helpers

    def load_plugins(self, plugins):
        # register plugins
        self.plugins.register_plugins(plugins)

        # load command plugins
        for Command in self.plugins.commands:
            command_obj = Command()
            self.commands.append(command_obj)

        # load device factory plugins
        for DeviceFactory in self.plugins.device_factories:
            type = DeviceFactory.type
            self.device_factories[type] = DeviceFactory()

    def open_devices(self, devices):
        for name, settings in devices.items():
            if 'type' not in settings:
                self.close_devices()
                message = f"type not found for device {name}"
                raise OpenDeviceError(message)

            # type, factory
            type    = settings.pop('type')
            factory = self.device_factories[type]

            # open and save
            device                  = factory.open(**settings)
            self.devices[name]      = device
            self.device_types[name] = type

    def close_devices(self):
        for name, device in self.devices.items():
            # type and factory
            type    = self.device_types[name]
            factory = self.device_factories[type]

            # close
            factory.close(device)

        # clear all devices
        self.devices.clear()
        self.device_types.clear()



from .helpers                 import is_plugin
from instrument_server.errors import PluginError
from importlib                import import_module


class CommandPluginManager:
    def __init__(self):
        self.plugins = []

    def register_plugin(self, module):
        # is plugin?
        if not is_plugin(module):
            return False

        plugin = module.plugin

        # already registered?
        if plugin in self.plugins:
            message = f"command plugin {plugin} is already registered"
            raise PluginError(message)

        # register
        self.plugins.append(plugin)
        return True



from setuptools import find_packages, setup


description='A tool for creating Test and Measurement microservices'


setup(
    name='instrument-server',
    version='2.1.0',
    description=description,
    long_description=description,
    author='Nick Lalic',
    author_email='nick.lalic@gmail.com',
    url='https://github.com/Terrabits/instrument-server',
    py_modules=['instrument_server'],
    packages=find_packages(exclude=['tests']),
    install_requires=['ruamel.yaml>=0.15.85<1.0', 'pyvisa>=1.9.1<2.0'],
    extras_require={
        'dev':  ['ddt>=1.2.0<2.0', 'twine>=3.8.0', 'wheel>=0.37.1'],
        'dist': ['twine>=3.8.0', 'wheel>=0.37.1'],
        'test': ['ddt>=1.2.0<2.0']
    },
    license='R&S Terms and Conditions for Royalty-Free Products',
    zip_safe=False,
    keywords='RF instrument SCPI',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Environment
        'Environment :: Console',

        # OS
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering',

        # Rohde & Schwarz Royalty-Free products license
        'License :: Other/Proprietary License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    entry_points={
    'console_scripts': [
        'instrument-server=instrument_server.bin.instrument_server:main'
    ]
    })



from .command_plugin_manager  import CommandPluginManager
from .device_factory_plugin_manager  import DeviceFactoryPluginManager
from .helpers                 import load_module
from instrument_server.errors import PluginError


class PluginManager:
    def __init__(self):
        self.commands_manager         = CommandPluginManager()
        self.device_factories_manager = DeviceFactoryPluginManager()

    def register_plugins(self, modules):
        for module in modules:
            self.register_plugin(module)

    def register_plugin(self, module):
        module = load_module(module)
        if self.commands_manager.register_plugin(module):
            # module is command plugin
            return
        if self.device_factories_manager.register_plugin(module):
            # module is device factory plugin
            return

        # module is not plugin
        message = f'module {module} is not recognized as a plugin'
        raise PluginError(message)

    @property
    def commands(self):
        return self.commands_manager.plugins

    @property
    def device_factories(self):
        return self.device_factories_manager.plugins



from   .helpers            import description_for, prog_for
from   ..instrument_server import InstrumentServer
from   ..yaml              import load_yaml
import argparse
import os
import sys


def main(app_name, plugins=None, devices=None):
    """Instrument Server Application Command Line Interface"""
    parser = argparse.ArgumentParser(prog=prog_for(app_name), description=description_for(app_name))
    parser.add_argument('--address', '-a', type=str, default='0.0.0.0', help='Listening address. Default: 0.0.0.0')
    parser.add_argument('--port',    '-p', type=int, default=9000,      help='Listening port. Default: 9000')
    if plugins is None:
        parser.add_argument('plugins_list_file', type=load_yaml, help='Plugins list file (yaml)')
    if devices is None:
        # require device file
        parser.add_argument('devices_file', type=load_yaml, help='Device connection file (yaml)')

    # parse
    args    = parser.parse_args()
    address = args.address
    port    = args.port
    plugins = plugins if plugins is not None else args.plugins_list_file
    devices = devices if devices is not None else args.devices_file

    # plugins list?
    if not isinstance(plugins, list):
        message = 'plugins must be list of modules'
        raise TypeError(message)

    # devices dict?
    if not isinstance(devices, dict):
        message = 'devices must be dict of device declarations'
        raise TypeError(message)

    # import plugins from the
    # current working directory
    sys.path.insert(0, os.getcwd())

    # run
    server = InstrumentServer(plugins, devices)
    server.run(address, port)
    sys.exit(0)



from .command_error           import CommandError
from .command_not_found_error import CommandNotFoundError
from .device_error            import DeviceError
from .open_device_error       import OpenDeviceError
from .plugin_error            import PluginError


# export
__all__ = [
    'CommandError',
    'CommandNotFoundError',
    'DeviceError',
    'OpenDeviceError',
    'PluginError'
]
